touch = 'touch'
temperature = 'temperature'
object_present = 'objectPresent'
humidity = 'humidity'
object_present_count = 'objectPresentCount'
touch_count = 'touchCount'
water_present = 'waterPresent'
battery_status = 'batteryStatus'
network_status = 'networkStatus'
labels_changed = 'labelsChanged'
connection_status = 'connectionStatus'
ethernet_status = 'ethernetStatus'
cellularStatus = 'cellularStatus'
